/*
 * Copyright (c) CERN 2013-2017
 *
 * Copyright (c) Members of the EMI Collaboration. 2011-2013
 *  See  http://www.eu-emi.eu/partners for details on the copyright
 *  holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <apr_strings.h>
#include <ctype.h>
#include <dmlite/c/catalog.h>
#include <dmlite/c/checksums.h>
#include <httpd.h>
#include <regex.h>
#include "shared/utils.h"
#include <http_log.h>
#include "mod_lcgdm_ns.h"

dav_error* dav_ns_digest_header(request_rec *r, const dav_resource *resource,
        char* output, size_t outsize)
{
    const char *want_digest;
    char digest_name[32], digest[1024], full_digest_name[64];
    int has_digest;

    want_digest = apr_table_get(r->headers_in, "Want-Digest");

    if (!want_digest) {
        return NULL;
    }

    has_digest = 0;
    while (dav_shared_next_digest(&want_digest, digest_name, sizeof(digest_name))) {
        snprintf(full_digest_name, sizeof(full_digest_name), "checksum.%s", digest_name);

        int ret = dmlite_getchecksum(resource->info->ctx,
            resource->info->sfn, // logical name
            full_digest_name, digest, sizeof(digest),
            NULL, // replica
            0, // 0 = do not recalculate
            0 // 1800 seconds wait, otherwise return EAGAIN
            );

        if (ret == DMLITE_RDR_ON_CHECKSUM) {
          // Beware, in this ugly case the digest var contains a destination url.
          // This happens because Dynafed/Ugr may like to redirect the checksum
          // request somewhere else, e.g. one of the endpoints or some external checksumming
          // machinery
          snprintf(output, outsize, "%s", digest);
          apr_table_set(resource->info->request->headers_out, "Location", digest);
          return dav_shared_new_error(resource->info->request, resource->info->ctx,
                                      HTTP_MOVED_TEMPORARILY, "=> %s", digest);
        }
        
        if (ret == 0 && digest[0] != '\0') {
          // Beware, the representation for some checksum types
          // has to be base64

          char digest_base64_value[1024];
          size_t digest_length = dav_shared_hexdigesttobase64(digest_name,
                                                              digest,
                                                              digest_base64_value);
          if (!digest_length) {
            return dav_shared_new_error(resource->info->request, resource->info->ctx,
                                        HTTP_INTERNAL_SERVER_ERROR, "Cannot handle digest '%s:%s'", full_digest_name, digest);
          }
            
          
          
          size_t len;
          len = snprintf(output, outsize, "%s=%s,", digest_name, digest_base64_value);
          output += len;
          outsize -= len;
          has_digest = 1;
        } else if (DMLITE_ERRNO(dmlite_errno(resource->info->ctx)) == EAGAIN) {
            return dav_shared_new_error(resource->info->request, resource->info->ctx,
                202, "Checksum not available yet. Try again later.");
        
        } else if (dmlite_errno(resource->info->ctx) != 0) {
            ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, resource->info->request,
                "Failed to get the checksum %s: %s", digest_name, dmlite_error(resource->info->ctx));
        } else {
            ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, resource->info->request,
                "Failed to get the checksum %s: empty value", digest_name);
        }
    }

    if (*(output-1) == ',') {
        --output;
    }
    *output = '\0';

    return NULL;
}
